@@ -1,20 +1,20 @@
-/*
- *
- * Copyright 2017 gRPC authors.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
- *
- *     http://www.apache.org/licenses/LICENSE-2.0
- *
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS,
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
- * See the License for the specific language governing permissions and
- * limitations under the License.
- *
- */
+//
+//
+// Copyright 2017 gRPC authors.
+//
+// Licensed under the Apache License, Version 2.0 (the "License");
+// you may not use this file except in compliance with the License.
+// You may obtain a copy of the License at
+//
+//     http://www.apache.org/licenses/LICENSE-2.0
+//
+// Unless required by applicable law or agreed to in writing, software
+// distributed under the License is distributed on an "AS IS" BASIS,
+// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+// See the License for the specific language governing permissions and
+// limitations under the License.
+//
+//
 
 #include <grpc/support/port_platform.h>
 
@@ -33,10 +33,9 @@
 #include "absl/strings/str_cat.h"
 #include "absl/strings/string_view.h"
 #include "absl/types/optional.h"
-#include "absl/utility/utility.h"
 
 #include <grpc/grpc.h>
-#include <grpc/impl/codegen/connectivity_state.h>
+#include <grpc/impl/connectivity_state.h>
 #include <grpc/status.h>
 #include <grpc/support/alloc.h>
 #include <grpc/support/log.h>
@@ -48,13 +47,13 @@
 #include "src/core/lib/config/core_configuration.h"
 #include "src/core/lib/gprpp/debug_location.h"
 #include "src/core/lib/gprpp/ref_counted_ptr.h"
+#include "src/core/lib/gprpp/status_helper.h"
 #include "src/core/lib/gprpp/time.h"
 #include "src/core/lib/iomgr/closure.h"
 #include "src/core/lib/iomgr/endpoint.h"
 #include "src/core/lib/iomgr/error.h"
 #include "src/core/lib/iomgr/exec_ctx.h"
 #include "src/core/lib/iomgr/iomgr_fwd.h"
-#include "src/core/lib/iomgr/pollset.h"
 #include "src/core/lib/resource_quota/arena.h"
 #include "src/core/lib/slice/slice.h"
 #include "src/core/lib/slice/slice_buffer.h"
@@ -86,7 +85,7 @@ void fill_in_metadata(inproc_stream* s, const grpc_metadata_batch* metadata,
                       grpc_metadata_batch* out_md, bool* markfilled);
 
 void ResetSendMessage(grpc_transport_stream_op_batch* batch) {
-  absl::exchange(batch->payload->send_message.send_message, nullptr)->Clear();
+  std::exchange(batch->payload->send_message.send_message, nullptr)->Clear();
 }
 
 struct shared_mu {
@@ -205,9 +204,9 @@ struct inproc_stream {
         cs->write_buffer_trailing_md.Clear();
         cs->write_buffer_trailing_md_filled = false;
       }
-      if (!GRPC_ERROR_IS_NONE(cs->write_buffer_cancel_error)) {
+      if (!cs->write_buffer_cancel_error.ok()) {
         cancel_other_error = cs->write_buffer_cancel_error;
-        cs->write_buffer_cancel_error = GRPC_ERROR_NONE;
+        cs->write_buffer_cancel_error = absl::OkStatus();
         maybe_process_ops_locked(this, cancel_other_error);
       }
 
@@ -215,13 +214,7 @@ struct inproc_stream {
     }
   }
 
-  ~inproc_stream() {
-    GRPC_ERROR_UNREF(write_buffer_cancel_error);
-    GRPC_ERROR_UNREF(cancel_self_error);
-    GRPC_ERROR_UNREF(cancel_other_error);
-
-    t->unref();
-  }
+  ~inproc_stream() { t->unref(); }
 
 #ifndef NDEBUG
 #define STREAM_REF(refs, reason) grpc_stream_ref(refs, reason)
@@ -259,7 +252,7 @@ struct inproc_stream {
       grpc_core::Timestamp::InfFuture();
   grpc_metadata_batch write_buffer_trailing_md{arena};
   bool write_buffer_trailing_md_filled = false;
-  grpc_error_handle write_buffer_cancel_error = GRPC_ERROR_NONE;
+  grpc_error_handle write_buffer_cancel_error;
 
   struct inproc_stream* other_side;
   bool other_side_closed = false;               // won't talk anymore
@@ -283,8 +276,8 @@ struct inproc_stream {
 
   bool closed = false;
 
-  grpc_error_handle cancel_self_error = GRPC_ERROR_NONE;
-  grpc_error_handle cancel_other_error = GRPC_ERROR_NONE;
+  grpc_error_handle cancel_self_error;
+  grpc_error_handle cancel_other_error;
 
   grpc_core::Timestamp deadline = grpc_core::Timestamp::InfFuture();
 
@@ -416,14 +409,13 @@ void complete_if_batch_end_locked(inproc_stream* s, grpc_error_handle error,
 
   if ((is_sm + is_stm + is_rim + is_rm + is_rtm) == 1) {
     INPROC_LOG(GPR_INFO, "%s %p %p %s", msg, s, op,
-               grpc_error_std_string(error).c_str());
-    grpc_core::ExecCtx::Run(DEBUG_LOCATION, op->on_complete,
-                            GRPC_ERROR_REF(error));
+               grpc_core::StatusToString(error).c_str());
+    grpc_core::ExecCtx::Run(DEBUG_LOCATION, op->on_complete, error);
   }
 }
 
 void maybe_process_ops_locked(inproc_stream* s, grpc_error_handle error) {
-  if (s && (!GRPC_ERROR_IS_NONE(error) || s->ops_needed)) {
+  if (s && (!error.ok() || s->ops_needed)) {
     s->ops_needed = false;
     op_state_machine_locked(s, error);
   }
@@ -447,12 +439,12 @@ void fail_helper_locked(inproc_stream* s, grpc_error_handle error) {
     fill_in_metadata(s, &fake_md, dest, destfilled);
 
     if (other != nullptr) {
-      if (GRPC_ERROR_IS_NONE(other->cancel_other_error)) {
-        other->cancel_other_error = GRPC_ERROR_REF(error);
+      if (other->cancel_other_error.ok()) {
+        other->cancel_other_error = error;
       }
       maybe_process_ops_locked(other, error);
-    } else if (GRPC_ERROR_IS_NONE(s->write_buffer_cancel_error)) {
-      s->write_buffer_cancel_error = GRPC_ERROR_REF(error);
+    } else if (s->write_buffer_cancel_error.ok()) {
+      s->write_buffer_cancel_error = error;
     }
   }
   if (s->recv_initial_md_op) {
@@ -470,9 +462,9 @@ void fail_helper_locked(inproc_stream* s, grpc_error_handle error) {
                        s->recv_initial_md_op->payload->recv_initial_metadata
                            .recv_initial_metadata,
                        nullptr);
-      err = GRPC_ERROR_NONE;
+      err = absl::OkStatus();
     } else {
-      err = GRPC_ERROR_REF(error);
+      err = error;
     }
     if (s->recv_initial_md_op->payload->recv_initial_metadata
             .trailing_metadata_available != nullptr) {
@@ -484,8 +476,8 @@ void fail_helper_locked(inproc_stream* s, grpc_error_handle error) {
     }
     INPROC_LOG(GPR_INFO,
                "fail_helper %p scheduling initial-metadata-ready %s %s", s,
-               grpc_error_std_string(error).c_str(),
-               grpc_error_std_string(err).c_str());
+               grpc_core::StatusToString(error).c_str(),
+               grpc_core::StatusToString(err).c_str());
     grpc_core::ExecCtx::Run(
         DEBUG_LOCATION,
         s->recv_initial_md_op->payload->recv_initial_metadata
@@ -500,7 +492,7 @@ void fail_helper_locked(inproc_stream* s, grpc_error_handle error) {
   }
   if (s->recv_message_op) {
     INPROC_LOG(GPR_INFO, "fail_helper %p scheduling message-ready %s", s,
-               grpc_error_std_string(error).c_str());
+               grpc_core::StatusToString(error).c_str());
     if (s->recv_message_op->payload->recv_message
             .call_failed_before_recv_message != nullptr) {
       *s->recv_message_op->payload->recv_message
@@ -508,8 +500,7 @@ void fail_helper_locked(inproc_stream* s, grpc_error_handle error) {
     }
     grpc_core::ExecCtx::Run(
         DEBUG_LOCATION,
-        s->recv_message_op->payload->recv_message.recv_message_ready,
-        GRPC_ERROR_REF(error));
+        s->recv_message_op->payload->recv_message.recv_message_ready, error);
     complete_if_batch_end_locked(
         s, error, s->recv_message_op,
         "fail_helper scheduling recv-message-on-complete");
@@ -530,14 +521,14 @@ void fail_helper_locked(inproc_stream* s, grpc_error_handle error) {
   }
   if (s->recv_trailing_md_op) {
     INPROC_LOG(GPR_INFO, "fail_helper %p scheduling trailing-metadata-ready %s",
-               s, grpc_error_std_string(error).c_str());
+               s, grpc_core::StatusToString(error).c_str());
     grpc_core::ExecCtx::Run(
         DEBUG_LOCATION,
         s->recv_trailing_md_op->payload->recv_trailing_metadata
             .recv_trailing_metadata_ready,
-        GRPC_ERROR_REF(error));
+        error);
     INPROC_LOG(GPR_INFO, "fail_helper %p scheduling trailing-md-on-complete %s",
-               s, grpc_error_std_string(error).c_str());
+               s, grpc_core::StatusToString(error).c_str());
     complete_if_batch_end_locked(
         s, error, s->recv_trailing_md_op,
         "fail_helper scheduling recv-trailing-metadata-on-complete");
@@ -545,8 +536,6 @@ void fail_helper_locked(inproc_stream* s, grpc_error_handle error) {
   }
   close_other_side_locked(s, "fail_helper:other_side");
   close_stream_locked(s);
-
-  GRPC_ERROR_UNREF(error);
 }
 
 // TODO(vjpai): It should not be necessary to drain the incoming byte
@@ -568,12 +557,12 @@ void message_transfer_locked(inproc_stream* sender, inproc_stream* receiver) {
   grpc_core::ExecCtx::Run(
       DEBUG_LOCATION,
       receiver->recv_message_op->payload->recv_message.recv_message_ready,
-      GRPC_ERROR_NONE);
+      absl::OkStatus());
   complete_if_batch_end_locked(
-      sender, GRPC_ERROR_NONE, sender->send_message_op,
+      sender, absl::OkStatus(), sender->send_message_op,
       "message_transfer scheduling sender on_complete");
   complete_if_batch_end_locked(
-      receiver, GRPC_ERROR_NONE, receiver->recv_message_op,
+      receiver, absl::OkStatus(), receiver->recv_message_op,
       "message_transfer scheduling receiver on_complete");
 
   receiver->recv_message_op = nullptr;
@@ -586,7 +575,7 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
   // Schedule our appropriate closures
   // and then return to ops_needed state if still needed
 
-  grpc_error_handle new_err = GRPC_ERROR_NONE;
+  grpc_error_handle new_err;
 
   bool needs_close = false;
 
@@ -594,26 +583,26 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
   // cancellation takes precedence
   inproc_stream* other = s->other_side;
 
-  if (!GRPC_ERROR_IS_NONE(s->cancel_self_error)) {
-    fail_helper_locked(s, GRPC_ERROR_REF(s->cancel_self_error));
+  if (!s->cancel_self_error.ok()) {
+    fail_helper_locked(s, s->cancel_self_error);
     goto done;
-  } else if (!GRPC_ERROR_IS_NONE(s->cancel_other_error)) {
-    fail_helper_locked(s, GRPC_ERROR_REF(s->cancel_other_error));
+  } else if (!s->cancel_other_error.ok()) {
+    fail_helper_locked(s, s->cancel_other_error);
     goto done;
-  } else if (!GRPC_ERROR_IS_NONE(error)) {
-    fail_helper_locked(s, GRPC_ERROR_REF(error));
+  } else if (!error.ok()) {
+    fail_helper_locked(s, error);
     goto done;
   }
 
   if (s->send_message_op && other) {
     if (other->recv_message_op) {
       message_transfer_locked(s, other);
-      maybe_process_ops_locked(other, GRPC_ERROR_NONE);
+      maybe_process_ops_locked(other, absl::OkStatus());
     } else if (!s->t->is_client && s->trailing_md_sent) {
       // A server send will never be matched if the server already sent status
       ResetSendMessage(s->send_message_op);
       complete_if_batch_end_locked(
-          s, GRPC_ERROR_NONE, s->send_message_op,
+          s, absl::OkStatus(), s->send_message_op,
           "op_state_machine scheduling send-message-on-complete case 1");
       s->send_message_op = nullptr;
     }
@@ -638,8 +627,8 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
     if (*destfilled || s->trailing_md_sent) {
       // The buffer is already in use; that's an error!
       INPROC_LOG(GPR_INFO, "Extra trailing metadata %p", s);
-      new_err = GRPC_ERROR_CREATE_FROM_STATIC_STRING("Extra trailing metadata");
-      fail_helper_locked(s, GRPC_ERROR_REF(new_err));
+      new_err = GRPC_ERROR_CREATE("Extra trailing metadata");
+      fail_helper_locked(s, new_err);
       goto done;
     } else {
       if (!other || !other->closed) {
@@ -659,32 +648,31 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
             DEBUG_LOCATION,
             s->recv_trailing_md_op->payload->recv_trailing_metadata
                 .recv_trailing_metadata_ready,
-            GRPC_ERROR_NONE);
+            absl::OkStatus());
         INPROC_LOG(GPR_INFO,
                    "op_state_machine %p scheduling trailing-md-on-complete", s);
         grpc_core::ExecCtx::Run(DEBUG_LOCATION,
                                 s->recv_trailing_md_op->on_complete,
-                                GRPC_ERROR_NONE);
+                                absl::OkStatus());
         s->recv_trailing_md_op = nullptr;
         needs_close = true;
       }
     }
-    maybe_process_ops_locked(other, GRPC_ERROR_NONE);
+    maybe_process_ops_locked(other, absl::OkStatus());
     complete_if_batch_end_locked(
-        s, GRPC_ERROR_NONE, s->send_trailing_md_op,
+        s, absl::OkStatus(), s->send_trailing_md_op,
         "op_state_machine scheduling send-trailing-metadata-on-complete");
     s->send_trailing_md_op = nullptr;
   }
   if (s->recv_initial_md_op) {
     if (s->initial_md_recvd) {
-      new_err =
-          GRPC_ERROR_CREATE_FROM_STATIC_STRING("Already recvd initial md");
+      new_err = GRPC_ERROR_CREATE("Already recvd initial md");
       INPROC_LOG(
           GPR_INFO,
           "op_state_machine %p scheduling on_complete errors for already "
           "recvd initial md %s",
-          s, grpc_error_std_string(new_err).c_str());
-      fail_helper_locked(s, GRPC_ERROR_REF(new_err));
+          s, grpc_core::StatusToString(new_err).c_str());
+      fail_helper_locked(s, new_err);
       goto done;
     }
 
@@ -711,9 +699,9 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
           DEBUG_LOCATION,
           s->recv_initial_md_op->payload->recv_initial_metadata
               .recv_initial_metadata_ready,
-          GRPC_ERROR_NONE);
+          absl::OkStatus());
       complete_if_batch_end_locked(
-          s, GRPC_ERROR_NONE, s->recv_initial_md_op,
+          s, absl::OkStatus(), s->recv_initial_md_op,
           "op_state_machine scheduling recv-initial-metadata-on-complete");
       s->recv_initial_md_op = nullptr;
     }
@@ -721,7 +709,7 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
   if (s->recv_message_op) {
     if (other && other->send_message_op) {
       message_transfer_locked(other, s);
-      maybe_process_ops_locked(other, GRPC_ERROR_NONE);
+      maybe_process_ops_locked(other, absl::OkStatus());
     }
   }
   if (s->to_read_trailing_md_filled) {
@@ -735,14 +723,13 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
         s->to_read_trailing_md_filled = false;
         s->trailing_md_recvd_implicit_only = false;
       } else {
-        new_err =
-            GRPC_ERROR_CREATE_FROM_STATIC_STRING("Already recvd trailing md");
+        new_err = GRPC_ERROR_CREATE("Already recvd trailing md");
         INPROC_LOG(
             GPR_INFO,
             "op_state_machine %p scheduling on_complete errors for already "
             "recvd trailing md %s",
-            s, grpc_error_std_string(new_err).c_str());
-        fail_helper_locked(s, GRPC_ERROR_REF(new_err));
+            s, grpc_core::StatusToString(new_err).c_str());
+        fail_helper_locked(s, new_err);
         goto done;
       }
     }
@@ -754,7 +741,7 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
       grpc_core::ExecCtx::Run(
           DEBUG_LOCATION,
           s->recv_message_op->payload->recv_message.recv_message_ready,
-          GRPC_ERROR_NONE);
+          absl::OkStatus());
       complete_if_batch_end_locked(
           s, new_err, s->recv_message_op,
           "op_state_machine scheduling recv-message-on-complete");
@@ -791,10 +778,10 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
             DEBUG_LOCATION,
             s->recv_trailing_md_op->payload->recv_trailing_metadata
                 .recv_trailing_metadata_ready,
-            GRPC_ERROR_NONE);
+            absl::OkStatus());
         grpc_core::ExecCtx::Run(DEBUG_LOCATION,
                                 s->recv_trailing_md_op->on_complete,
-                                GRPC_ERROR_NONE);
+                                absl::OkStatus());
         s->recv_trailing_md_op = nullptr;
         needs_close = s->trailing_md_sent;
       }
@@ -810,12 +797,12 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
     // because we have already sent status and the RPC is over as far as we
     // are concerned.
     INPROC_LOG(GPR_INFO, "op_state_machine %p scheduling trailing-md-ready %s",
-               s, grpc_error_std_string(new_err).c_str());
+               s, grpc_core::StatusToString(new_err).c_str());
     grpc_core::ExecCtx::Run(
         DEBUG_LOCATION,
         s->recv_trailing_md_op->payload->recv_trailing_metadata
             .recv_trailing_metadata_ready,
-        GRPC_ERROR_REF(new_err));
+        new_err);
     complete_if_batch_end_locked(
         s, new_err, s->recv_trailing_md_op,
         "op_state_machine scheduling recv-trailing-md-on-complete");
@@ -833,7 +820,7 @@ void op_state_machine_locked(inproc_stream* s, grpc_error_handle error) {
     grpc_core::ExecCtx::Run(
         DEBUG_LOCATION,
         s->recv_message_op->payload->recv_message.recv_message_ready,
-        GRPC_ERROR_NONE);
+        absl::OkStatus());
     complete_if_batch_end_locked(
         s, new_err, s->recv_message_op,
         "op_state_machine scheduling recv-message-on-complete");
@@ -863,16 +850,15 @@ done:
     close_other_side_locked(s, "op_state_machine");
     close_stream_locked(s);
   }
-  GRPC_ERROR_UNREF(new_err);
 }
 
 bool cancel_stream_locked(inproc_stream* s, grpc_error_handle error) {
   bool ret = false;  // was the cancel accepted
   INPROC_LOG(GPR_INFO, "cancel_stream %p with %s", s,
-             grpc_error_std_string(error).c_str());
-  if (GRPC_ERROR_IS_NONE(s->cancel_self_error)) {
+             grpc_core::StatusToString(error).c_str());
+  if (s->cancel_self_error.ok()) {
     ret = true;
-    s->cancel_self_error = GRPC_ERROR_REF(error);
+    s->cancel_self_error = error;
     // Catch current value of other before it gets closed off
     inproc_stream* other = s->other_side;
     maybe_process_ops_locked(s, s->cancel_self_error);
@@ -890,12 +876,12 @@ bool cancel_stream_locked(inproc_stream* s, grpc_error_handle error) {
     fill_in_metadata(s, &cancel_md, dest, destfilled);
 
     if (other != nullptr) {
-      if (GRPC_ERROR_IS_NONE(other->cancel_other_error)) {
-        other->cancel_other_error = GRPC_ERROR_REF(s->cancel_self_error);
+      if (other->cancel_other_error.ok()) {
+        other->cancel_other_error = s->cancel_self_error;
       }
       maybe_process_ops_locked(other, other->cancel_other_error);
-    } else if (GRPC_ERROR_IS_NONE(s->write_buffer_cancel_error)) {
-      s->write_buffer_cancel_error = GRPC_ERROR_REF(s->cancel_self_error);
+    } else if (s->write_buffer_cancel_error.ok()) {
+      s->write_buffer_cancel_error = s->cancel_self_error;
     }
 
     // if we are a server and already received trailing md but
@@ -906,7 +892,7 @@ bool cancel_stream_locked(inproc_stream* s, grpc_error_handle error) {
           DEBUG_LOCATION,
           s->recv_trailing_md_op->payload->recv_trailing_metadata
               .recv_trailing_metadata_ready,
-          GRPC_ERROR_REF(s->cancel_self_error));
+          s->cancel_self_error);
       complete_if_batch_end_locked(
           s, s->cancel_self_error, s->recv_trailing_md_op,
           "cancel_stream scheduling trailing-md-on-complete");
@@ -917,7 +903,6 @@ bool cancel_stream_locked(inproc_stream* s, grpc_error_handle error) {
   close_other_side_locked(s, "cancel_stream:other_side");
   close_stream_locked(s);
 
-  GRPC_ERROR_UNREF(error);
   return ret;
 }
 
@@ -940,7 +925,7 @@ void perform_stream_op(grpc_transport* gt, grpc_stream* gs,
                    s->t->is_client, false);
     }
   }
-  grpc_error_handle error = GRPC_ERROR_NONE;
+  grpc_error_handle error;
   grpc_closure* on_complete = op->on_complete;
   // TODO(roth): This is a hack needed because we use data inside of the
   // closure itself to do the barrier calculation (i.e., to ensure that
@@ -957,9 +942,9 @@ void perform_stream_op(grpc_transport* gt, grpc_stream* gs,
     // this function is responsible to make sure that that field gets unref'ed
     cancel_stream_locked(s, op->payload->cancel_stream.cancel_error);
     // this op can complete without an error
-  } else if (!GRPC_ERROR_IS_NONE(s->cancel_self_error)) {
+  } else if (!s->cancel_self_error.ok()) {
     // already self-canceled so still give it an error
-    error = GRPC_ERROR_REF(s->cancel_self_error);
+    error = s->cancel_self_error;
   } else {
     INPROC_LOG(GPR_INFO, "perform_stream_op %p %s%s%s%s%s%s%s", s,
                s->t->is_client ? "client" : "server",
@@ -972,12 +957,11 @@ void perform_stream_op(grpc_transport* gt, grpc_stream* gs,
   }
 
   inproc_stream* other = s->other_side;
-  if (GRPC_ERROR_IS_NONE(error) &&
-      (op->send_initial_metadata || op->send_trailing_metadata)) {
+  if (error.ok() && (op->send_initial_metadata || op->send_trailing_metadata)) {
     if (s->t->is_closed) {
-      error = GRPC_ERROR_CREATE_FROM_STATIC_STRING("Endpoint already shutdown");
+      error = GRPC_ERROR_CREATE("Endpoint already shutdown");
     }
-    if (GRPC_ERROR_IS_NONE(error) && op->send_initial_metadata) {
+    if (error.ok() && op->send_initial_metadata) {
       grpc_metadata_batch* dest = (other == nullptr)
                                       ? &s->write_buffer_initial_md
                                       : &other->to_read_initial_md;
@@ -986,7 +970,7 @@ void perform_stream_op(grpc_transport* gt, grpc_stream* gs,
       if (*destfilled || s->initial_md_sent) {
         // The buffer is already in use; that's an error!
         INPROC_LOG(GPR_INFO, "Extra initial metadata %p", s);
-        error = GRPC_ERROR_CREATE_FROM_STATIC_STRING("Extra initial metadata");
+        error = GRPC_ERROR_CREATE("Extra initial metadata");
       } else {
         if (!s->other_side_closed) {
           fill_in_metadata(
@@ -1007,10 +991,9 @@ void perform_stream_op(grpc_transport* gt, grpc_stream* gs,
     }
   }
 
-  if (GRPC_ERROR_IS_NONE(error) &&
-      (op->send_message || op->send_trailing_metadata ||
-       op->recv_initial_metadata || op->recv_message ||
-       op->recv_trailing_metadata)) {
+  if (error.ok() && (op->send_message || op->send_trailing_metadata ||
+                     op->recv_initial_metadata || op->recv_message ||
+                     op->recv_trailing_metadata)) {
     // Mark ops that need to be processed by the state machine
     if (op->send_message) {
       s->send_message_op = op;
@@ -1047,7 +1030,7 @@ void perform_stream_op(grpc_transport* gt, grpc_stream* gs,
       s->ops_needed = true;
     }
   } else {
-    if (!GRPC_ERROR_IS_NONE(error)) {
+    if (!error.ok()) {
       // Consume any send message that was sent here but that we are not
       // pushing to the other side
       if (op->send_message) {
@@ -1067,42 +1050,41 @@ void perform_stream_op(grpc_transport* gt, grpc_stream* gs,
         INPROC_LOG(
             GPR_INFO,
             "perform_stream_op error %p scheduling initial-metadata-ready %s",
-            s, grpc_error_std_string(error).c_str());
+            s, grpc_core::StatusToString(error).c_str());
         grpc_core::ExecCtx::Run(
             DEBUG_LOCATION,
             op->payload->recv_initial_metadata.recv_initial_metadata_ready,
-            GRPC_ERROR_REF(error));
+            error);
       }
       if (op->recv_message) {
         INPROC_LOG(
             GPR_INFO,
             "perform_stream_op error %p scheduling recv message-ready %s", s,
-            grpc_error_std_string(error).c_str());
+            grpc_core::StatusToString(error).c_str());
         if (op->payload->recv_message.call_failed_before_recv_message !=
             nullptr) {
           *op->payload->recv_message.call_failed_before_recv_message = true;
         }
         grpc_core::ExecCtx::Run(DEBUG_LOCATION,
                                 op->payload->recv_message.recv_message_ready,
-                                GRPC_ERROR_REF(error));
+                                error);
       }
       if (op->recv_trailing_metadata) {
         INPROC_LOG(GPR_INFO,
                    "perform_stream_op error %p scheduling "
                    "trailing-metadata-ready %s",
-                   s, grpc_error_std_string(error).c_str());
+                   s, grpc_core::StatusToString(error).c_str());
         grpc_core::ExecCtx::Run(
             DEBUG_LOCATION,
             op->payload->recv_trailing_metadata.recv_trailing_metadata_ready,
-            GRPC_ERROR_REF(error));
+            error);
       }
     }
     INPROC_LOG(GPR_INFO, "perform_stream_op %p scheduling on_complete %s", s,
-               grpc_error_std_string(error).c_str());
-    grpc_core::ExecCtx::Run(DEBUG_LOCATION, on_complete, GRPC_ERROR_REF(error));
+               grpc_core::StatusToString(error).c_str());
+    grpc_core::ExecCtx::Run(DEBUG_LOCATION, on_complete, error);
   }
   gpr_mu_unlock(mu);
-  GRPC_ERROR_UNREF(error);
 }
 
 void close_transport_locked(inproc_transport* t) {
@@ -1111,14 +1093,14 @@ void close_transport_locked(inproc_transport* t) {
                             "close transport");
   if (!t->is_closed) {
     t->is_closed = true;
-    /* Also end all streams on this transport */
+    // Also end all streams on this transport
     while (t->stream_list != nullptr) {
       // cancel_stream_locked also adjusts stream list
       cancel_stream_locked(
           t->stream_list,
-          grpc_error_set_int(
-              GRPC_ERROR_CREATE_FROM_STATIC_STRING("Transport closed"),
-              GRPC_ERROR_INT_GRPC_STATUS, GRPC_STATUS_UNAVAILABLE));
+          grpc_error_set_int(GRPC_ERROR_CREATE("Transport closed"),
+                             grpc_core::StatusIntProperty::kRpcStatus,
+                             GRPC_STATUS_UNAVAILABLE));
     }
   }
 }
@@ -1139,17 +1121,15 @@ void perform_transport_op(grpc_transport* gt, grpc_transport_op* op) {
     t->accept_stream_data = op->set_accept_stream_user_data;
   }
   if (op->on_consumed) {
-    grpc_core::ExecCtx::Run(DEBUG_LOCATION, op->on_consumed, GRPC_ERROR_NONE);
+    grpc_core::ExecCtx::Run(DEBUG_LOCATION, op->on_consumed, absl::OkStatus());
   }
 
   bool do_close = false;
-  if (!GRPC_ERROR_IS_NONE(op->goaway_error)) {
+  if (!op->goaway_error.ok()) {
     do_close = true;
-    GRPC_ERROR_UNREF(op->goaway_error);
   }
-  if (!GRPC_ERROR_IS_NONE(op->disconnect_with_error)) {
+  if (!op->disconnect_with_error.ok()) {
     do_close = true;
-    GRPC_ERROR_UNREF(op->disconnect_with_error);
   }
 
   if (do_close) {
@@ -1168,7 +1148,7 @@ void destroy_stream(grpc_transport* gt, grpc_stream* gs,
   gpr_mu_unlock(&t->mu->mu);
   s->~inproc_stream();
   grpc_core::ExecCtx::Run(DEBUG_LOCATION, then_schedule_closure,
-                          GRPC_ERROR_NONE);
+                          absl::OkStatus());
 }
 
 void destroy_transport(grpc_transport* gt) {
@@ -1181,9 +1161,9 @@ void destroy_transport(grpc_transport* gt) {
   t->unref();
 }
 
-/*******************************************************************************
- * INTEGRATION GLUE
- */
+//******************************************************************************
+// INTEGRATION GLUE
+//
 
 void set_pollset(grpc_transport* /*gt*/, grpc_stream* /*gs*/,
                  grpc_pollset* /*pollset*/) {
@@ -1205,9 +1185,9 @@ const grpc_transport_vtable inproc_vtable = {
     destroy_stream,        destroy_transport,
     get_endpoint};
 
-/*******************************************************************************
- * Main inproc transport functions
- */
+//******************************************************************************
+// Main inproc transport functions
+//
 void inproc_transports_create(grpc_transport** server_transport,
                               grpc_transport** client_transport) {
   INPROC_LOG(GPR_INFO, "inproc_transports_create");
@@ -1253,19 +1233,19 @@ grpc_channel* grpc_inproc_channel_create(grpc_server* server,
   grpc_error_handle error = core_server->SetupTransport(
       server_transport, nullptr, server_args, nullptr);
   grpc_channel* channel = nullptr;
-  if (GRPC_ERROR_IS_NONE(error)) {
+  if (error.ok()) {
     auto new_channel = grpc_core::Channel::Create(
         "inproc", client_args, GRPC_CLIENT_DIRECT_CHANNEL, client_transport);
     if (!new_channel.ok()) {
       GPR_ASSERT(!channel);
       gpr_log(GPR_ERROR, "Failed to create client channel: %s",
-              grpc_error_std_string(error).c_str());
+              grpc_core::StatusToString(error).c_str());
       intptr_t integer;
       grpc_status_code status = GRPC_STATUS_INTERNAL;
-      if (grpc_error_get_int(error, GRPC_ERROR_INT_GRPC_STATUS, &integer)) {
+      if (grpc_error_get_int(error, grpc_core::StatusIntProperty::kRpcStatus,
+                             &integer)) {
         status = static_cast<grpc_status_code>(integer);
       }
-      GRPC_ERROR_UNREF(error);
       // client_transport was destroyed when grpc_channel_create_internal saw an
       // error.
       grpc_transport_destroy(server_transport);
@@ -1277,13 +1257,13 @@ grpc_channel* grpc_inproc_channel_create(grpc_server* server,
   } else {
     GPR_ASSERT(!channel);
     gpr_log(GPR_ERROR, "Failed to create server channel: %s",
-            grpc_error_std_string(error).c_str());
+            grpc_core::StatusToString(error).c_str());
     intptr_t integer;
     grpc_status_code status = GRPC_STATUS_INTERNAL;
-    if (grpc_error_get_int(error, GRPC_ERROR_INT_GRPC_STATUS, &integer)) {
+    if (grpc_error_get_int(error, grpc_core::StatusIntProperty::kRpcStatus,
+                           &integer)) {
       status = static_cast<grpc_status_code>(integer);
     }
-    GRPC_ERROR_UNREF(error);
     grpc_transport_destroy(client_transport);
     grpc_transport_destroy(server_transport);
     channel = grpc_lame_client_channel_create(
