@@ -1,39 +1,37 @@
-/*
- *
- * Copyright 2017 gRPC authors.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
- *
- *     http://www.apache.org/licenses/LICENSE-2.0
- *
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS,
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
- * See the License for the specific language governing permissions and
- * limitations under the License.
- *
- */
+//
+//
+// Copyright 2017 gRPC authors.
+//
+// Licensed under the Apache License, Version 2.0 (the "License");
+// you may not use this file except in compliance with the License.
+// You may obtain a copy of the License at
+//
+//     http://www.apache.org/licenses/LICENSE-2.0
+//
+// Unless required by applicable law or agreed to in writing, software
+// distributed under the License is distributed on an "AS IS" BASIS,
+// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+// See the License for the specific language governing permissions and
+// limitations under the License.
+//
+//
 
 #include <grpc/support/port_platform.h>
 
 #include "src/core/lib/gprpp/fork.h"
 
-#include <string.h>
-
-#include <grpc/support/alloc.h>
+#include <grpc/support/atm.h>
 #include <grpc/support/sync.h>
 #include <grpc/support/time.h>
 
-#include "src/core/lib/gpr/useful.h"
-#include "src/core/lib/gprpp/global_config.h"
-#include "src/core/lib/gprpp/memory.h"
+#include "src/core/lib/event_engine/thread_local.h"
+#include "src/core/lib/gprpp/global_config_env.h"
+#include "src/core/lib/gprpp/no_destruct.h"
 
-/*
- * NOTE: FORKING IS NOT GENERALLY SUPPORTED, THIS IS ONLY INTENDED TO WORK
- *       AROUND VERY SPECIFIC USE CASES.
- */
+//
+// NOTE: FORKING IS NOT GENERALLY SUPPORTED, THIS IS ONLY INTENDED TO WORK
+//       AROUND VERY SPECIFIC USE CASES.
+//
 
 #ifdef GRPC_ENABLE_FORK_SUPPORT
 #define GRPC_ENABLE_FORK_SUPPORT_DEFAULT true
@@ -43,10 +41,10 @@
 
 GPR_GLOBAL_CONFIG_DEFINE_BOOL(grpc_enable_fork_support,
                               GRPC_ENABLE_FORK_SUPPORT_DEFAULT,
-                              "Enable folk support");
+                              "Enable fork support");
 
 namespace grpc_core {
-namespace internal {
+namespace {
 // The exec_ctx_count has 2 modes, blocked and unblocked.
 // When unblocked, the count is 2-indexed; exec_ctx_count=2 indicates
 // 0 active ExecCtxs, exex_ctx_count=3 indicates 1 active ExecCtxs...
@@ -54,7 +52,7 @@ namespace internal {
 // When blocked, the exec_ctx_count is 0-indexed.  Note that ExecCtx
 // creation can only be blocked if there is exactly 1 outstanding ExecCtx,
 // meaning that BLOCKED and UNBLOCKED counts partition the integers
-#define UNBLOCKED(n) (n + 2)
+#define UNBLOCKED(n) ((n) + 2)
 #define BLOCKED(n) (n)
 
 class ExecCtxState {
@@ -66,6 +64,12 @@ class ExecCtxState {
   }
 
   void IncExecCtxCount() {
+    // EventEngine is expected to terminate all threads before fork, and so this
+    // extra work is unnecessary
+    if (grpc_event_engine::experimental::ThreadLocal::IsEventEngineThread()) {
+      gpr_atm_no_barrier_fetch_add(&count_, 1);
+      return;
+    }
     gpr_atm count = gpr_atm_no_barrier_load(&count_);
     while (true) {
       if (count <= BLOCKED(1)) {
@@ -168,75 +172,78 @@ class ThreadState {
 
 void Fork::GlobalInit() {
   if (!override_enabled_) {
-    support_enabled_.Store(GPR_GLOBAL_CONFIG_GET(grpc_enable_fork_support),
-                           MemoryOrder::RELAXED);
-  }
-  if (support_enabled_.Load(MemoryOrder::RELAXED)) {
-    exec_ctx_state_ = grpc_core::New<internal::ExecCtxState>();
-    thread_state_ = grpc_core::New<internal::ThreadState>();
+    support_enabled_.store(GPR_GLOBAL_CONFIG_GET(grpc_enable_fork_support),
+                           std::memory_order_relaxed);
   }
 }
 
-void Fork::GlobalShutdown() {
-  if (support_enabled_.Load(MemoryOrder::RELAXED)) {
-    grpc_core::Delete(exec_ctx_state_);
-    grpc_core::Delete(thread_state_);
-  }
+bool Fork::Enabled() {
+  return support_enabled_.load(std::memory_order_relaxed);
 }
 
-bool Fork::Enabled() { return support_enabled_.Load(MemoryOrder::RELAXED); }
-
 // Testing Only
 void Fork::Enable(bool enable) {
   override_enabled_ = true;
-  support_enabled_.Store(enable, MemoryOrder::RELAXED);
+  support_enabled_.store(enable, std::memory_order_relaxed);
 }
 
-void Fork::DoIncExecCtxCount() { exec_ctx_state_->IncExecCtxCount(); }
+void Fork::DoIncExecCtxCount() {
+  NoDestructSingleton<ExecCtxState>::Get()->IncExecCtxCount();
+}
 
-void Fork::DoDecExecCtxCount() { exec_ctx_state_->DecExecCtxCount(); }
+void Fork::DoDecExecCtxCount() {
+  NoDestructSingleton<ExecCtxState>::Get()->DecExecCtxCount();
+}
 
 void Fork::SetResetChildPollingEngineFunc(
     Fork::child_postfork_func reset_child_polling_engine) {
-  reset_child_polling_engine_ = reset_child_polling_engine;
+  if (reset_child_polling_engine_ == nullptr) {
+    reset_child_polling_engine_ = new std::vector<Fork::child_postfork_func>();
+  }
+  if (reset_child_polling_engine == nullptr) {
+    reset_child_polling_engine_->clear();
+  } else {
+    reset_child_polling_engine_->emplace_back(reset_child_polling_engine);
+  }
 }
-Fork::child_postfork_func Fork::GetResetChildPollingEngineFunc() {
-  return reset_child_polling_engine_;
+
+const std::vector<Fork::child_postfork_func>&
+Fork::GetResetChildPollingEngineFunc() {
+  return *reset_child_polling_engine_;
 }
 
 bool Fork::BlockExecCtx() {
-  if (support_enabled_.Load(MemoryOrder::RELAXED)) {
-    return exec_ctx_state_->BlockExecCtx();
+  if (support_enabled_.load(std::memory_order_relaxed)) {
+    return NoDestructSingleton<ExecCtxState>::Get()->BlockExecCtx();
   }
   return false;
 }
 
 void Fork::AllowExecCtx() {
-  if (support_enabled_.Load(MemoryOrder::RELAXED)) {
-    exec_ctx_state_->AllowExecCtx();
+  if (support_enabled_.load(std::memory_order_relaxed)) {
+    NoDestructSingleton<ExecCtxState>::Get()->AllowExecCtx();
   }
 }
 
 void Fork::IncThreadCount() {
-  if (support_enabled_.Load(MemoryOrder::RELAXED)) {
-    thread_state_->IncThreadCount();
+  if (support_enabled_.load(std::memory_order_relaxed)) {
+    NoDestructSingleton<ThreadState>::Get()->IncThreadCount();
   }
 }
 
 void Fork::DecThreadCount() {
-  if (support_enabled_.Load(MemoryOrder::RELAXED)) {
-    thread_state_->DecThreadCount();
+  if (support_enabled_.load(std::memory_order_relaxed)) {
+    NoDestructSingleton<ThreadState>::Get()->DecThreadCount();
   }
 }
 void Fork::AwaitThreads() {
-  if (support_enabled_.Load(MemoryOrder::RELAXED)) {
-    thread_state_->AwaitThreads();
+  if (support_enabled_.load(std::memory_order_relaxed)) {
+    NoDestructSingleton<ThreadState>::Get()->AwaitThreads();
   }
 }
 
-internal::ExecCtxState* Fork::exec_ctx_state_ = nullptr;
-internal::ThreadState* Fork::thread_state_ = nullptr;
-Atomic<bool> Fork::support_enabled_(false);
+std::atomic<bool> Fork::support_enabled_(false);
 bool Fork::override_enabled_ = false;
-Fork::child_postfork_func Fork::reset_child_polling_engine_ = nullptr;
+std::vector<Fork::child_postfork_func>* Fork::reset_child_polling_engine_ =
+    nullptr;
 }  // namespace grpc_core
