@@ -12,8 +12,8 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-#ifndef GRPC_CORE_LIB_GPRPP_TIME_H
-#define GRPC_CORE_LIB_GPRPP_TIME_H
+#ifndef GRPC_SRC_CORE_LIB_GPRPP_TIME_H
+#define GRPC_SRC_CORE_LIB_GPRPP_TIME_H
 
 #include <grpc/support/port_platform.h>
 
@@ -363,4 +363,4 @@ std::ostream& operator<<(std::ostream& out, Duration duration);
 
 }  // namespace grpc_core
 
-#endif  // GRPC_CORE_LIB_GPRPP_TIME_H
+#endif  // GRPC_SRC_CORE_LIB_GPRPP_TIME_H
