@@ -1,22 +1,22 @@
-/*
- *
- * Copyright 2015 gRPC authors.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
- *
- *     http://www.apache.org/licenses/LICENSE-2.0
- *
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS,
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
- * See the License for the specific language governing permissions and
- * limitations under the License.
- *
- */
+//
+//
+// Copyright 2015 gRPC authors.
+//
+// Licensed under the Apache License, Version 2.0 (the "License");
+// you may not use this file except in compliance with the License.
+// You may obtain a copy of the License at
+//
+//     http://www.apache.org/licenses/LICENSE-2.0
+//
+// Unless required by applicable law or agreed to in writing, software
+// distributed under the License is distributed on an "AS IS" BASIS,
+// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+// See the License for the specific language governing permissions and
+// limitations under the License.
+//
+//
 
-/* FIXME: "posix" files shouldn't be depending on _GNU_SOURCE */
+// FIXME: "posix" files shouldn't be depending on _GNU_SOURCE
 #ifndef _GNU_SOURCE
 #define _GNU_SOURCE
 #endif
@@ -29,6 +29,7 @@
 
 #include <errno.h>
 #include <fcntl.h>
+#include <inttypes.h>
 #include <netinet/in.h>
 #include <netinet/tcp.h>
 #include <string.h>
@@ -42,50 +43,177 @@
 #include "absl/strings/str_cat.h"
 #include "absl/strings/str_format.h"
 
+#include <grpc/byte_buffer.h>
+#include <grpc/event_engine/endpoint_config.h>
 #include <grpc/support/alloc.h>
 #include <grpc/support/log.h>
 #include <grpc/support/sync.h>
 #include <grpc/support/time.h>
 
 #include "src/core/lib/address_utils/sockaddr_utils.h"
-#include "src/core/lib/channel/channel_args.h"
+#include "src/core/lib/event_engine/default_event_engine.h"
+#include "src/core/lib/event_engine/resolved_address_internal.h"
+#include "src/core/lib/event_engine/shim.h"
+#include "src/core/lib/event_engine/tcp_socket_utils.h"
 #include "src/core/lib/gpr/string.h"
+#include "src/core/lib/gprpp/crash.h"
 #include "src/core/lib/gprpp/memory.h"
+#include "src/core/lib/gprpp/strerror.h"
+#include "src/core/lib/iomgr/event_engine_shims/closure.h"
+#include "src/core/lib/iomgr/event_engine_shims/endpoint.h"
 #include "src/core/lib/iomgr/exec_ctx.h"
 #include "src/core/lib/iomgr/resolve_address.h"
 #include "src/core/lib/iomgr/sockaddr.h"
 #include "src/core/lib/iomgr/socket_utils_posix.h"
+#include "src/core/lib/iomgr/systemd_utils.h"
 #include "src/core/lib/iomgr/tcp_posix.h"
 #include "src/core/lib/iomgr/tcp_server.h"
 #include "src/core/lib/iomgr/tcp_server_utils_posix.h"
 #include "src/core/lib/iomgr/unix_sockets_posix.h"
+#include "src/core/lib/resource_quota/api.h"
+#include "src/core/lib/transport/error_utils.h"
+
+static std::atomic<int64_t> num_dropped_connections{0};
+
+using ::grpc_event_engine::experimental::EndpointConfig;
+using ::grpc_event_engine::experimental::EventEngine;
+using ::grpc_event_engine::experimental::MemoryAllocator;
+using ::grpc_event_engine::experimental::PosixEventEngineWithFdSupport;
+using ::grpc_event_engine::experimental::SliceBuffer;
+
+class MemoryAllocatorFactoryWrapper
+    : public grpc_event_engine::experimental::MemoryAllocatorFactory {
+ public:
+  explicit MemoryAllocatorFactoryWrapper(
+      grpc_core::MemoryQuotaRefPtr memory_quota)
+      : memory_quota_(std::move(memory_quota)) {}
+
+  MemoryAllocator CreateMemoryAllocator(absl::string_view name) override {
+    return memory_quota_->CreateMemoryAllocator(name);
+  }
+
+ private:
+  grpc_core::MemoryQuotaRefPtr memory_quota_;
+};
+
+static grpc_error_handle CreateEventEngineListener(
+    grpc_tcp_server* s, grpc_closure* shutdown_complete,
+    const EndpointConfig& config, grpc_tcp_server** server) {
+  PosixEventEngineWithFdSupport::PosixAcceptCallback accept_cb =
+      [s](int listener_fd, std::unique_ptr<EventEngine::Endpoint> ep,
+          bool is_external, MemoryAllocator /*allocator*/,
+          SliceBuffer* pending_data) {
+        grpc_core::ApplicationCallbackExecCtx app_ctx;
+        grpc_core::ExecCtx exec_ctx;
+        grpc_tcp_server_acceptor* acceptor =
+            static_cast<grpc_tcp_server_acceptor*>(
+                gpr_malloc(sizeof(*acceptor)));
+        acceptor->from_server = s;
+        acceptor->port_index = -1;
+        acceptor->fd_index = -1;
+        if (!is_external) {
+          auto it = s->listen_fd_to_index_map.find(listener_fd);
+          if (it != s->listen_fd_to_index_map.end()) {
+            acceptor->port_index = std::get<0>(it->second);
+            acceptor->fd_index = std::get<1>(it->second);
+          }
+        } else {
+          // External connection handling.
+          grpc_resolved_address addr;
+          memset(&addr, 0, sizeof(addr));
+          addr.len = static_cast<socklen_t>(sizeof(struct sockaddr_storage));
+          // Get the fd of the socket connected to peer.
+          int fd =
+              reinterpret_cast<
+                  grpc_event_engine::experimental::PosixEndpointWithFdSupport*>(
+                  ep.get())
+                  ->GetWrappedFd();
+          if (getpeername(fd, reinterpret_cast<struct sockaddr*>(addr.addr),
+                          &(addr.len)) < 0) {
+            gpr_log(GPR_ERROR, "Failed getpeername: %s",
+                    grpc_core::StrError(errno).c_str());
+            close(fd);
+            return;
+          }
+          (void)grpc_set_socket_no_sigpipe_if_possible(fd);
+          auto addr_uri = grpc_sockaddr_to_uri(&addr);
+          if (!addr_uri.ok()) {
+            gpr_log(GPR_ERROR, "Invalid address: %s",
+                    addr_uri.status().ToString().c_str());
+            return;
+          }
+          if (grpc_tcp_trace.enabled()) {
+            gpr_log(GPR_INFO,
+                    "SERVER_CONNECT: incoming external connection: %s",
+                    addr_uri->c_str());
+          }
+        }
+        grpc_pollset* read_notifier_pollset =
+            (*(s->pollsets))[static_cast<size_t>(gpr_atm_no_barrier_fetch_add(
+                                 &s->next_pollset_to_assign, 1)) %
+                             s->pollsets->size()];
+        acceptor->external_connection = is_external;
+        acceptor->listener_fd = listener_fd;
+        grpc_byte_buffer* buf = nullptr;
+        if (pending_data != nullptr && pending_data->Length() > 0) {
+          buf = grpc_raw_byte_buffer_create(nullptr, 0);
+          grpc_slice_buffer_swap(&buf->data.raw.slice_buffer,
+                                 pending_data->c_slice_buffer());
+          pending_data->Clear();
+        }
+        acceptor->pending_data = buf;
+        s->on_accept_cb(
+            s->on_accept_cb_arg,
+            grpc_event_engine::experimental::grpc_event_engine_endpoint_create(
+                std::move(ep)),
+            read_notifier_pollset, acceptor);
+      };
+  auto on_shutdown_complete_cb =
+      grpc_event_engine::experimental::GrpcClosureToStatusCallback(
+          shutdown_complete);
+  PosixEventEngineWithFdSupport* engine_ptr =
+      reinterpret_cast<PosixEventEngineWithFdSupport*>(
+          config.GetVoidPointer(GRPC_INTERNAL_ARG_EVENT_ENGINE));
+  // Keeps the engine alive for some tests that have not otherwise instantiated
+  // an EventEngine
+  std::shared_ptr<EventEngine> keeper;
+  if (engine_ptr == nullptr) {
+    keeper = grpc_event_engine::experimental::GetDefaultEventEngine();
+    engine_ptr = reinterpret_cast<PosixEventEngineWithFdSupport*>(keeper.get());
+  }
+  auto listener = engine_ptr->CreatePosixListener(
+      std::move(accept_cb),
+      [s, shutdown_complete](absl::Status status) {
+        grpc_event_engine::experimental::RunEventEngineClosure(
+            shutdown_complete, absl_status_to_grpc_error(status));
+        delete s->fd_handler;
+        delete s;
+      },
+      config, std::make_unique<MemoryAllocatorFactoryWrapper>(s->memory_quota));
+  if (!listener.ok()) {
+    delete s;
+    *server = nullptr;
+    return listener.status();
+  }
+  s->ee_listener = std::move(*listener);
+  return absl::OkStatus();
+}
 
 static grpc_error_handle tcp_server_create(grpc_closure* shutdown_complete,
-                                           const grpc_channel_args* args,
+                                           const EndpointConfig& config,
+                                           grpc_tcp_server_cb on_accept_cb,
+                                           void* on_accept_cb_arg,
                                            grpc_tcp_server** server) {
-  grpc_tcp_server* s =
-      static_cast<grpc_tcp_server*>(gpr_zalloc(sizeof(grpc_tcp_server)));
+  grpc_tcp_server* s = new grpc_tcp_server;
   s->so_reuseport = grpc_is_socket_reuse_port_supported();
   s->expand_wildcard_addrs = false;
-  for (size_t i = 0; i < (args == nullptr ? 0 : args->num_args); i++) {
-    if (0 == strcmp(GRPC_ARG_ALLOW_REUSEPORT, args->args[i].key)) {
-      if (args->args[i].type == GRPC_ARG_INTEGER) {
-        s->so_reuseport = grpc_is_socket_reuse_port_supported() &&
-                          (args->args[i].value.integer != 0);
-      } else {
-        gpr_free(s);
-        return GRPC_ERROR_CREATE_FROM_STATIC_STRING(GRPC_ARG_ALLOW_REUSEPORT
-                                                    " must be an integer");
-      }
-    } else if (0 == strcmp(GRPC_ARG_EXPAND_WILDCARD_ADDRS, args->args[i].key)) {
-      if (args->args[i].type == GRPC_ARG_INTEGER) {
-        s->expand_wildcard_addrs = (args->args[i].value.integer != 0);
-      } else {
-        gpr_free(s);
-        return GRPC_ERROR_CREATE_FROM_STATIC_STRING(
-            GRPC_ARG_EXPAND_WILDCARD_ADDRS " must be an integer");
-      }
-    }
+  auto value = config.GetInt(GRPC_ARG_ALLOW_REUSEPORT);
+  if (value.has_value()) {
+    s->so_reuseport = (grpc_is_socket_reuse_port_supported() && *value != 0);
+  }
+  value = config.GetInt(GRPC_ARG_EXPAND_WILDCARD_ADDRS);
+  if (value.has_value()) {
+    s->expand_wildcard_addrs = (*value != 0);
   }
   gpr_ref_init(&s->refs, 1);
   gpr_mu_init(&s->mu);
@@ -94,17 +222,31 @@ static grpc_error_handle tcp_server_create(grpc_closure* shutdown_complete,
   s->shutdown = false;
   s->shutdown_starting.head = nullptr;
   s->shutdown_starting.tail = nullptr;
-  s->shutdown_complete = shutdown_complete;
-  s->on_accept_cb = nullptr;
-  s->on_accept_cb_arg = nullptr;
+  if (!grpc_event_engine::experimental::UseEventEngineListener()) {
+    s->shutdown_complete = shutdown_complete;
+  } else {
+    s->shutdown_complete = nullptr;
+  }
+  s->on_accept_cb = on_accept_cb;
+  s->on_accept_cb_arg = on_accept_cb_arg;
   s->head = nullptr;
   s->tail = nullptr;
   s->nports = 0;
-  s->channel_args = grpc_channel_args_copy(args);
+  s->options = TcpOptionsFromEndpointConfig(config);
   s->fd_handler = nullptr;
+  GPR_ASSERT(s->options.resource_quota != nullptr);
+  GPR_ASSERT(s->on_accept_cb);
+  s->memory_quota = s->options.resource_quota->memory_quota();
+  s->pre_allocated_fd = -1;
   gpr_atm_no_barrier_store(&s->next_pollset_to_assign, 0);
+  s->n_bind_ports = 0;
+  new (&s->listen_fd_to_index_map)
+      absl::flat_hash_map<int, std::tuple<int, int>>();
   *server = s;
-  return GRPC_ERROR_NONE;
+  if (grpc_event_engine::experimental::UseEventEngineListener()) {
+    return CreateEventEngineListener(s, shutdown_complete, config, server);
+  }
+  return absl::OkStatus();
 }
 
 static void finish_shutdown(grpc_tcp_server* s) {
@@ -113,20 +255,22 @@ static void finish_shutdown(grpc_tcp_server* s) {
   gpr_mu_unlock(&s->mu);
   if (s->shutdown_complete != nullptr) {
     grpc_core::ExecCtx::Run(DEBUG_LOCATION, s->shutdown_complete,
-                            GRPC_ERROR_NONE);
+                            absl::OkStatus());
   }
-
   gpr_mu_destroy(&s->mu);
-
   while (s->head) {
     grpc_tcp_listener* sp = s->head;
     s->head = sp->next;
     gpr_free(sp);
   }
-  grpc_channel_args_destroy(s->channel_args);
-  delete s->fd_handler;
-
-  gpr_free(s);
+  if (grpc_event_engine::experimental::UseEventEngineListener()) {
+    // This will trigger asynchronous execution of the on_shutdown_complete
+    // callback when appropriate. That callback will delete the server
+    s->ee_listener.reset();
+  } else {
+    delete s->fd_handler;
+    delete s;
+  }
 }
 
 static void destroyed_port(void* server, grpc_error_handle /*error*/) {
@@ -142,11 +286,11 @@ static void destroyed_port(void* server, grpc_error_handle /*error*/) {
   }
 }
 
-/* called when all listening endpoints have been shutdown, so no further
-   events will be received on them - at this point it's safe to destroy
-   things */
+// called when all listening endpoints have been shutdown, so no further
+// events will be received on them - at this point it's safe to destroy
+// things
 static void deactivated_all_ports(grpc_tcp_server* s) {
-  /* delete ALL the things */
+  // delete ALL the things
   gpr_mu_lock(&s->mu);
 
   GPR_ASSERT(s->shutdown);
@@ -154,7 +298,10 @@ static void deactivated_all_ports(grpc_tcp_server* s) {
   if (s->head) {
     grpc_tcp_listener* sp;
     for (sp = s->head; sp; sp = sp->next) {
-      grpc_unlink_if_unix_domain_socket(&sp->addr);
+      // Do not unlink if there is a pre-allocated FD
+      if (grpc_tcp_server_pre_allocated_fd(s) <= 0) {
+        grpc_unlink_if_unix_domain_socket(&sp->addr);
+      }
       GRPC_CLOSURE_INIT(&sp->destroyed_closure, destroyed_port, s,
                         grpc_schedule_on_exec_ctx);
       grpc_fd_orphan(sp->emfd, &sp->destroyed_closure, nullptr,
@@ -169,16 +316,13 @@ static void deactivated_all_ports(grpc_tcp_server* s) {
 
 static void tcp_server_destroy(grpc_tcp_server* s) {
   gpr_mu_lock(&s->mu);
-
   GPR_ASSERT(!s->shutdown);
   s->shutdown = true;
-
-  /* shutdown all fd's */
+  // shutdown all fd's
   if (s->active_ports) {
     grpc_tcp_listener* sp;
     for (sp = s->head; sp; sp = sp->next) {
-      grpc_fd_shutdown(
-          sp->emfd, GRPC_ERROR_CREATE_FROM_STATIC_STRING("Server destroyed"));
+      grpc_fd_shutdown(sp->emfd, GRPC_ERROR_CREATE("Server destroyed"));
     }
     gpr_mu_unlock(&s->mu);
   } else {
@@ -187,70 +331,90 @@ static void tcp_server_destroy(grpc_tcp_server* s) {
   }
 }
 
-/* event manager callback when reads are ready */
+// event manager callback when reads are ready
 static void on_read(void* arg, grpc_error_handle err) {
   grpc_tcp_listener* sp = static_cast<grpc_tcp_listener*>(arg);
   grpc_pollset* read_notifier_pollset;
-  if (err != GRPC_ERROR_NONE) {
+  if (!err.ok()) {
     goto error;
   }
 
-  /* loop until accept4 returns EAGAIN, and then re-arm notification */
+  // loop until accept4 returns EAGAIN, and then re-arm notification
   for (;;) {
     grpc_resolved_address addr;
     memset(&addr, 0, sizeof(addr));
     addr.len = static_cast<socklen_t>(sizeof(struct sockaddr_storage));
-    /* Note: If we ever decide to return this address to the user, remember to
-       strip off the ::ffff:0.0.0.0/96 prefix first. */
+    // Note: If we ever decide to return this address to the user, remember to
+    // strip off the ::ffff:0.0.0.0/96 prefix first.
     int fd = grpc_accept4(sp->fd, &addr, 1, 1);
     if (fd < 0) {
-      switch (errno) {
-        case EINTR:
-          continue;
-        case EAGAIN:
-          grpc_fd_notify_on_read(sp->emfd, &sp->read_closure);
-          return;
-        default:
-          gpr_mu_lock(&sp->server->mu);
-          if (!sp->server->shutdown_listeners) {
-            gpr_log(GPR_ERROR, "Failed accept4: %s", strerror(errno));
-          } else {
-            /* if we have shutdown listeners, accept4 could fail, and we
-               needn't notify users */
-          }
-          gpr_mu_unlock(&sp->server->mu);
-          goto error;
+      if (errno == EINTR) {
+        continue;
+      } else if (errno == EAGAIN || errno == ECONNABORTED ||
+                 errno == EWOULDBLOCK) {
+        grpc_fd_notify_on_read(sp->emfd, &sp->read_closure);
+        return;
+      } else {
+        gpr_mu_lock(&sp->server->mu);
+        if (!sp->server->shutdown_listeners) {
+          gpr_log(GPR_ERROR, "Failed accept4: %s",
+                  grpc_core::StrError(errno).c_str());
+        } else {
+          // if we have shutdown listeners, accept4 could fail, and we
+          // needn't notify users
+        }
+        gpr_mu_unlock(&sp->server->mu);
+        goto error;
       }
     }
 
-    /* For UNIX sockets, the accept call might not fill up the member sun_path
-     * of sockaddr_un, so explicitly call getsockname to get it. */
+    if (sp->server->memory_quota->IsMemoryPressureHigh()) {
+      int64_t dropped_connections_count =
+          num_dropped_connections.fetch_add(1, std::memory_order_relaxed) + 1;
+      if (dropped_connections_count % 1000 == 1) {
+        gpr_log(GPR_INFO,
+                "Dropped >= %" PRId64
+                " new connection attempts due to high memory pressure",
+                dropped_connections_count);
+      }
+      close(fd);
+      continue;
+    }
+
+    // For UNIX sockets, the accept call might not fill up the member sun_path
+    // of sockaddr_un, so explicitly call getpeername to get it.
     if (grpc_is_unix_socket(&addr)) {
       memset(&addr, 0, sizeof(addr));
       addr.len = static_cast<socklen_t>(sizeof(struct sockaddr_storage));
-      if (getsockname(fd, reinterpret_cast<struct sockaddr*>(addr.addr),
+      if (getpeername(fd, reinterpret_cast<struct sockaddr*>(addr.addr),
                       &(addr.len)) < 0) {
-        gpr_log(GPR_ERROR, "Failed getsockname: %s", strerror(errno));
+        gpr_log(GPR_ERROR, "Failed getpeername: %s",
+                grpc_core::StrError(errno).c_str());
         close(fd);
         goto error;
       }
     }
 
-    grpc_set_socket_no_sigpipe_if_possible(fd);
+    (void)grpc_set_socket_no_sigpipe_if_possible(fd);
 
     err = grpc_apply_socket_mutator_in_args(fd, GRPC_FD_SERVER_CONNECTION_USAGE,
-                                            sp->server->channel_args);
-    if (err != GRPC_ERROR_NONE) {
+                                            sp->server->options);
+    if (!err.ok()) {
       goto error;
     }
 
-    std::string addr_str = grpc_sockaddr_to_uri(&addr);
+    auto addr_uri = grpc_sockaddr_to_uri(&addr);
+    if (!addr_uri.ok()) {
+      gpr_log(GPR_ERROR, "Invalid address: %s",
+              addr_uri.status().ToString().c_str());
+      goto error;
+    }
     if (GRPC_TRACE_FLAG_ENABLED(grpc_tcp_trace)) {
       gpr_log(GPR_INFO, "SERVER_CONNECT: incoming connection: %s",
-              addr_str.c_str());
+              addr_uri->c_str());
     }
 
-    std::string name = absl::StrCat("tcp-server-connection:", addr_str);
+    std::string name = absl::StrCat("tcp-server-connection:", addr_uri.value());
     grpc_fd* fdobj = grpc_fd_create(fd, name.c_str(), true);
 
     read_notifier_pollset = (*(sp->server->pollsets))
@@ -267,14 +431,13 @@ static void on_read(void* arg, grpc_error_handle err) {
     acceptor->port_index = sp->port_index;
     acceptor->fd_index = sp->fd_index;
     acceptor->external_connection = false;
-
     sp->server->on_accept_cb(
         sp->server->on_accept_cb_arg,
-        grpc_tcp_create(fdobj, sp->server->channel_args, addr_str.c_str()),
+        grpc_tcp_create(fdobj, sp->server->options, addr_uri.value()),
         read_notifier_pollset, acceptor);
   }
 
-  GPR_UNREACHABLE_CODE(return );
+  GPR_UNREACHABLE_CODE(return);
 
 error:
   gpr_mu_lock(&sp->server->mu);
@@ -286,7 +449,7 @@ error:
   }
 }
 
-/* Treat :: or 0.0.0.0 as a family-agnostic wildcard. */
+// Treat :: or 0.0.0.0 as a family-agnostic wildcard.
 static grpc_error_handle add_wildcard_addrs_to_server(grpc_tcp_server* s,
                                                       unsigned port_index,
                                                       int requested_port,
@@ -297,8 +460,8 @@ static grpc_error_handle add_wildcard_addrs_to_server(grpc_tcp_server* s,
   grpc_dualstack_mode dsmode;
   grpc_tcp_listener* sp = nullptr;
   grpc_tcp_listener* sp2 = nullptr;
-  grpc_error_handle v6_err = GRPC_ERROR_NONE;
-  grpc_error_handle v4_err = GRPC_ERROR_NONE;
+  grpc_error_handle v6_err;
+  grpc_error_handle v4_err;
   *out_port = -1;
 
   if (grpc_tcp_server_have_ifaddrs() && s->expand_wildcard_addrs) {
@@ -307,19 +470,19 @@ static grpc_error_handle add_wildcard_addrs_to_server(grpc_tcp_server* s,
   }
 
   grpc_sockaddr_make_wildcards(requested_port, &wild4, &wild6);
-  /* Try listening on IPv6 first. */
+  // Try listening on IPv6 first.
   if ((v6_err = grpc_tcp_server_add_addr(s, &wild6, port_index, fd_index,
-                                         &dsmode, &sp)) == GRPC_ERROR_NONE) {
+                                         &dsmode, &sp)) == absl::OkStatus()) {
     ++fd_index;
     requested_port = *out_port = sp->port;
     if (dsmode == GRPC_DSMODE_DUALSTACK || dsmode == GRPC_DSMODE_IPV4) {
-      return GRPC_ERROR_NONE;
+      return absl::OkStatus();
     }
   }
-  /* If we got a v6-only socket or nothing, try adding 0.0.0.0. */
+  // If we got a v6-only socket or nothing, try adding 0.0.0.0.
   grpc_sockaddr_set_port(&wild4, requested_port);
   if ((v4_err = grpc_tcp_server_add_addr(s, &wild4, port_index, fd_index,
-                                         &dsmode, &sp2)) == GRPC_ERROR_NONE) {
+                                         &dsmode, &sp2)) == absl::OkStatus()) {
     *out_port = sp2->port;
     if (sp != nullptr) {
       sp2->is_sibling = 1;
@@ -327,25 +490,23 @@ static grpc_error_handle add_wildcard_addrs_to_server(grpc_tcp_server* s,
     }
   }
   if (*out_port > 0) {
-    if (v6_err != GRPC_ERROR_NONE) {
+    if (!v6_err.ok()) {
       gpr_log(GPR_INFO,
               "Failed to add :: listener, "
               "the environment may not support IPv6: %s",
-              grpc_error_std_string(v6_err).c_str());
-      GRPC_ERROR_UNREF(v6_err);
+              grpc_core::StatusToString(v6_err).c_str());
     }
-    if (v4_err != GRPC_ERROR_NONE) {
+    if (!v4_err.ok()) {
       gpr_log(GPR_INFO,
               "Failed to add 0.0.0.0 listener, "
               "the environment may not support IPv4: %s",
-              grpc_error_std_string(v4_err).c_str());
-      GRPC_ERROR_UNREF(v4_err);
+              grpc_core::StatusToString(v4_err).c_str());
     }
-    return GRPC_ERROR_NONE;
+    return absl::OkStatus();
   } else {
-    grpc_error_handle root_err = GRPC_ERROR_CREATE_FROM_STATIC_STRING(
-        "Failed to add any wildcard listeners");
-    GPR_ASSERT(v6_err != GRPC_ERROR_NONE && v4_err != GRPC_ERROR_NONE);
+    grpc_error_handle root_err =
+        GRPC_ERROR_CREATE("Failed to add any wildcard listeners");
+    GPR_ASSERT(!v6_err.ok() && !v4_err.ok());
     root_err = grpc_error_add_child(root_err, v6_err);
     root_err = grpc_error_add_child(root_err, v4_err);
     return root_err;
@@ -355,7 +516,7 @@ static grpc_error_handle add_wildcard_addrs_to_server(grpc_tcp_server* s,
 static grpc_error_handle clone_port(grpc_tcp_listener* listener,
                                     unsigned count) {
   grpc_tcp_listener* sp = nullptr;
-  std::string addr_str;
+  absl::StatusOr<std::string> addr_str;
   grpc_error_handle err;
 
   for (grpc_tcp_listener* l = listener->next; l && l->is_sibling; l = l->next) {
@@ -368,17 +529,20 @@ static grpc_error_handle clone_port(grpc_tcp_listener* listener,
     grpc_dualstack_mode dsmode;
     err = grpc_create_dualstack_socket(&listener->addr, SOCK_STREAM, 0, &dsmode,
                                        &fd);
-    if (err != GRPC_ERROR_NONE) return err;
+    if (!err.ok()) return err;
     err = grpc_tcp_server_prepare_socket(listener->server, fd, &listener->addr,
                                          true, &port);
-    if (err != GRPC_ERROR_NONE) return err;
+    if (!err.ok()) return err;
     listener->server->nports++;
     addr_str = grpc_sockaddr_to_string(&listener->addr, true);
+    if (!addr_str.ok()) {
+      return GRPC_ERROR_CREATE(addr_str.status().ToString());
+    }
     sp = static_cast<grpc_tcp_listener*>(gpr_malloc(sizeof(grpc_tcp_listener)));
     sp->next = listener->next;
     listener->next = sp;
-    /* sp (the new listener) is a sibling of 'listener' (the original
-       listener). */
+    // sp (the new listener) is a sibling of 'listener' (the original
+    // listener).
     sp->is_sibling = 1;
     sp->sibling = listener->sibling;
     listener->sibling = sp;
@@ -386,7 +550,7 @@ static grpc_error_handle clone_port(grpc_tcp_listener* listener,
     sp->fd = fd;
     sp->emfd = grpc_fd_create(
         fd,
-        absl::StrFormat("tcp-server-listener:%s/clone-%d", addr_str.c_str(), i)
+        absl::StrFormat("tcp-server-listener:%s/clone-%d", *addr_str, i)
             .c_str(),
         true);
     memcpy(&sp->addr, &listener->addr, sizeof(grpc_resolved_address));
@@ -399,12 +563,36 @@ static grpc_error_handle clone_port(grpc_tcp_listener* listener,
     }
   }
 
-  return GRPC_ERROR_NONE;
+  return absl::OkStatus();
 }
 
 static grpc_error_handle tcp_server_add_port(grpc_tcp_server* s,
                                              const grpc_resolved_address* addr,
                                              int* out_port) {
+  if (grpc_event_engine::experimental::UseEventEngineListener()) {
+    gpr_mu_lock(&s->mu);
+    if (s->shutdown_listeners) {
+      gpr_mu_unlock(&s->mu);
+      return absl::UnknownError("Server already shutdown");
+    }
+    int fd_index = 0;
+    auto port = s->ee_listener->BindWithFd(
+        grpc_event_engine::experimental::CreateResolvedAddress(*addr),
+        [s, &fd_index](absl::StatusOr<int> listen_fd) {
+          if (!listen_fd.ok()) {
+            return;
+          }
+          GPR_DEBUG_ASSERT(*listen_fd > 0);
+          s->listen_fd_to_index_map.insert_or_assign(
+              *listen_fd, std::make_tuple(s->n_bind_ports, fd_index++));
+        });
+    if (port.ok()) {
+      s->n_bind_ports++;
+      *out_port = *port;
+    }
+    gpr_mu_unlock(&s->mu);
+    return port.status();
+  }
   GPR_ASSERT(addr->len <= GRPC_MAX_SOCKADDR_SIZE);
   grpc_tcp_listener* sp;
   grpc_resolved_address sockname_temp;
@@ -417,10 +605,9 @@ static grpc_error_handle tcp_server_add_port(grpc_tcp_server* s,
   if (s->tail != nullptr) {
     port_index = s->tail->port_index + 1;
   }
-  grpc_unlink_if_unix_domain_socket(addr);
 
-  /* Check if this is a wildcard port, and if so, try to keep the port the same
-     as some previously created listener. */
+  // Check if this is a wildcard port, and if so, try to keep the port the same
+  // as some previously created listener.
   if (requested_port == 0) {
     for (sp = s->head; sp; sp = sp->next) {
       sockname_temp.len =
@@ -440,6 +627,16 @@ static grpc_error_handle tcp_server_add_port(grpc_tcp_server* s,
       }
     }
   }
+
+  /* Check if systemd has pre-allocated valid FDs */
+  set_matching_sd_fds(s, addr, requested_port);
+
+  /* Do not unlink if there are pre-allocated FDs, or it will stop
+     working after the first client connects */
+  if (grpc_tcp_server_pre_allocated_fd(s) <= 0) {
+    grpc_unlink_if_unix_domain_socket(addr);
+  }
+
   if (grpc_sockaddr_is_wildcard(addr, &requested_port)) {
     return add_wildcard_addrs_to_server(s, port_index, requested_port,
                                         out_port);
@@ -448,14 +645,14 @@ static grpc_error_handle tcp_server_add_port(grpc_tcp_server* s,
     addr = &addr6_v4mapped;
   }
   if ((err = grpc_tcp_server_add_addr(s, addr, port_index, 0, &dsmode, &sp)) ==
-      GRPC_ERROR_NONE) {
+      absl::OkStatus()) {
     *out_port = sp->port;
   }
   return err;
 }
 
-/* Return listener at port_index or NULL. Should only be called with s->mu
-   locked. */
+// Return listener at port_index or NULL. Should only be called with s->mu
+// locked.
 static grpc_tcp_listener* get_port_index(grpc_tcp_server* s,
                                          unsigned port_index) {
   unsigned num_ports = 0;
@@ -473,6 +670,17 @@ static grpc_tcp_listener* get_port_index(grpc_tcp_server* s,
 unsigned tcp_server_port_fd_count(grpc_tcp_server* s, unsigned port_index) {
   unsigned num_fds = 0;
   gpr_mu_lock(&s->mu);
+  if (grpc_event_engine::experimental::UseEventEngineListener()) {
+    // This doesn't need to be very fast. Used in tests.
+    for (auto it = s->listen_fd_to_index_map.begin();
+         it != s->listen_fd_to_index_map.end(); it++) {
+      if (std::get<0>(it->second) == static_cast<int>(port_index)) {
+        num_fds++;
+      }
+    }
+    gpr_mu_unlock(&s->mu);
+    return num_fds;
+  }
   grpc_tcp_listener* sp = get_port_index(s, port_index);
   for (; sp; sp = sp->sibling) {
     ++num_fds;
@@ -484,6 +692,19 @@ unsigned tcp_server_port_fd_count(grpc_tcp_server* s, unsigned port_index) {
 static int tcp_server_port_fd(grpc_tcp_server* s, unsigned port_index,
                               unsigned fd_index) {
   gpr_mu_lock(&s->mu);
+  if (grpc_event_engine::experimental::UseEventEngineListener()) {
+    // This doesn't need to be very fast. Used in tests.
+    for (auto it = s->listen_fd_to_index_map.begin();
+         it != s->listen_fd_to_index_map.end(); it++) {
+      if (std::get<0>(it->second) == static_cast<int>(port_index) &&
+          std::get<1>(it->second) == static_cast<int>(fd_index)) {
+        gpr_mu_unlock(&s->mu);
+        return it->first;
+      }
+    }
+    gpr_mu_unlock(&s->mu);
+    return -1;
+  }
   grpc_tcp_listener* sp = get_port_index(s, port_index);
   for (; sp; sp = sp->sibling, --fd_index) {
     if (fd_index == 0) {
@@ -496,18 +717,19 @@ static int tcp_server_port_fd(grpc_tcp_server* s, unsigned port_index,
 }
 
 static void tcp_server_start(grpc_tcp_server* s,
-                             const std::vector<grpc_pollset*>* pollsets,
-                             grpc_tcp_server_cb on_accept_cb,
-                             void* on_accept_cb_arg) {
+                             const std::vector<grpc_pollset*>* pollsets) {
   size_t i;
   grpc_tcp_listener* sp;
-  GPR_ASSERT(on_accept_cb);
   gpr_mu_lock(&s->mu);
-  GPR_ASSERT(!s->on_accept_cb);
+  GPR_ASSERT(s->on_accept_cb);
   GPR_ASSERT(s->active_ports == 0);
-  s->on_accept_cb = on_accept_cb;
-  s->on_accept_cb_arg = on_accept_cb_arg;
   s->pollsets = pollsets;
+  if (grpc_event_engine::experimental::UseEventEngineListener()) {
+    GPR_ASSERT(!s->shutdown_listeners);
+    GPR_ASSERT(GRPC_LOG_IF_ERROR("listener_start", s->ee_listener->Start()));
+    gpr_mu_unlock(&s->mu);
+    return;
+  }
   sp = s->head;
   while (sp != nullptr) {
     if (s->so_reuseport && !grpc_is_unix_socket(&sp->addr) &&
@@ -545,7 +767,7 @@ static void tcp_server_shutdown_starting_add(grpc_tcp_server* s,
                                              grpc_closure* shutdown_starting) {
   gpr_mu_lock(&s->mu);
   grpc_closure_list_append(&s->shutdown_starting, shutdown_starting,
-                           GRPC_ERROR_NONE);
+                           absl::OkStatus());
   gpr_mu_unlock(&s->mu);
 }
 
@@ -562,17 +784,29 @@ static void tcp_server_unref(grpc_tcp_server* s) {
 static void tcp_server_shutdown_listeners(grpc_tcp_server* s) {
   gpr_mu_lock(&s->mu);
   s->shutdown_listeners = true;
+  if (grpc_event_engine::experimental::UseEventEngineListener()) {
+    s->ee_listener->ShutdownListeningFds();
+  }
   /* shutdown all fd's */
   if (s->active_ports) {
     grpc_tcp_listener* sp;
     for (sp = s->head; sp; sp = sp->next) {
-      grpc_fd_shutdown(sp->emfd,
-                       GRPC_ERROR_CREATE_FROM_STATIC_STRING("Server shutdown"));
+      grpc_fd_shutdown(sp->emfd, GRPC_ERROR_CREATE("Server shutdown"));
     }
   }
   gpr_mu_unlock(&s->mu);
 }
 
+static int tcp_server_pre_allocated_fd(grpc_tcp_server* s) {
+  return s->pre_allocated_fd;
+}
+
+static void tcp_server_set_pre_allocated_fd(grpc_tcp_server* s, int fd) {
+  gpr_mu_lock(&s->mu);
+  s->pre_allocated_fd = fd;
+  gpr_mu_unlock(&s->mu);
+}
+
 namespace {
 class ExternalConnectionHandler : public grpc_core::TcpServerFdHandler {
  public:
@@ -580,6 +814,18 @@ class ExternalConnectionHandler : public grpc_core::TcpServerFdHandler {
 
   // TODO(yangg) resolve duplicate code with on_read
   void Handle(int listener_fd, int fd, grpc_byte_buffer* buf) override {
+    if (grpc_event_engine::experimental::UseEventEngineListener()) {
+      grpc_event_engine::experimental::SliceBuffer pending_data;
+      if (buf != nullptr) {
+        pending_data =
+            grpc_event_engine::experimental::SliceBuffer::TakeCSliceBuffer(
+                buf->data.raw.slice_buffer);
+      }
+      GPR_ASSERT(GRPC_LOG_IF_ERROR("listener_handle_external_connection",
+                                   s_->ee_listener->HandleExternalConnection(
+                                       listener_fd, fd, &pending_data)));
+      return;
+    }
     grpc_pollset* read_notifier_pollset;
     grpc_resolved_address addr;
     memset(&addr, 0, sizeof(addr));
@@ -588,17 +834,23 @@ class ExternalConnectionHandler : public grpc_core::TcpServerFdHandler {
 
     if (getpeername(fd, reinterpret_cast<struct sockaddr*>(addr.addr),
                     &(addr.len)) < 0) {
-      gpr_log(GPR_ERROR, "Failed getpeername: %s", strerror(errno));
+      gpr_log(GPR_ERROR, "Failed getpeername: %s",
+              grpc_core::StrError(errno).c_str());
       close(fd);
       return;
     }
-    grpc_set_socket_no_sigpipe_if_possible(fd);
-    std::string addr_str = grpc_sockaddr_to_uri(&addr);
+    (void)grpc_set_socket_no_sigpipe_if_possible(fd);
+    auto addr_uri = grpc_sockaddr_to_uri(&addr);
+    if (!addr_uri.ok()) {
+      gpr_log(GPR_ERROR, "Invalid address: %s",
+              addr_uri.status().ToString().c_str());
+      return;
+    }
     if (grpc_tcp_trace.enabled()) {
       gpr_log(GPR_INFO, "SERVER_CONNECT: incoming external connection: %s",
-              addr_str.c_str());
+              addr_uri->c_str());
     }
-    std::string name = absl::StrCat("tcp-server-connection:", addr_str);
+    std::string name = absl::StrCat("tcp-server-connection:", addr_uri.value());
     grpc_fd* fdobj = grpc_fd_create(fd, name.c_str(), true);
     read_notifier_pollset =
         (*(s_->pollsets))[static_cast<size_t>(gpr_atm_no_barrier_fetch_add(
@@ -614,7 +866,7 @@ class ExternalConnectionHandler : public grpc_core::TcpServerFdHandler {
     acceptor->listener_fd = listener_fd;
     acceptor->pending_data = buf;
     s_->on_accept_cb(s_->on_accept_cb_arg,
-                     grpc_tcp_create(fdobj, s_->channel_args, addr_str.c_str()),
+                     grpc_tcp_create(fdobj, s_->options, addr_uri.value()),
                      read_notifier_pollset, acceptor);
   }
 
@@ -630,10 +882,17 @@ static grpc_core::TcpServerFdHandler* tcp_server_create_fd_handler(
 }
 
 grpc_tcp_server_vtable grpc_posix_tcp_server_vtable = {
-    tcp_server_create,        tcp_server_start,
-    tcp_server_add_port,      tcp_server_create_fd_handler,
-    tcp_server_port_fd_count, tcp_server_port_fd,
-    tcp_server_ref,           tcp_server_shutdown_starting_add,
-    tcp_server_unref,         tcp_server_shutdown_listeners};
+    tcp_server_create,
+    tcp_server_start,
+    tcp_server_add_port,
+    tcp_server_create_fd_handler,
+    tcp_server_port_fd_count,
+    tcp_server_port_fd,
+    tcp_server_ref,
+    tcp_server_shutdown_starting_add,
+    tcp_server_unref,
+    tcp_server_shutdown_listeners,
+    tcp_server_pre_allocated_fd,
+    tcp_server_set_pre_allocated_fd};
 
-#endif /* GRPC_POSIX_SOCKET_TCP_SERVER */
+#endif  // GRPC_POSIX_SOCKET_TCP_SERVER
