@@ -32,7 +32,6 @@
 #include "src/core/lib/event_engine/default_event_engine.h"
 #include "src/core/lib/gpr/alloc.h"
 #include "src/core/lib/iomgr/exec_ctx.h"
-#include "src/core/lib/promise/context.h"
 #include "src/core/lib/slice/slice.h"
 #include "src/core/lib/transport/transport_impl.h"
 
@@ -172,8 +171,6 @@ void grpc_transport_stream_op_batch_finish_with_failure(
 void grpc_transport_stream_op_batch_queue_finish_with_failure(
     grpc_transport_stream_op_batch* batch, grpc_error_handle error,
     grpc_core::CallCombinerClosureList* closures) {
-  if (batch->cancel_stream) {
-  }
   // Construct a list of closures to execute.
   if (batch->recv_initial_metadata) {
     closures->Add(
@@ -196,8 +193,6 @@ void grpc_transport_stream_op_batch_queue_finish_with_failure(
 
 void grpc_transport_stream_op_batch_finish_with_failure_from_transport(
     grpc_transport_stream_op_batch* batch, grpc_error_handle error) {
-  if (batch->cancel_stream) {
-  }
   // Construct a list of closures to execute.
   if (batch->recv_initial_metadata) {
     grpc_core::ExecCtx::Run(
@@ -273,9 +268,9 @@ grpc_transport_stream_op_batch* grpc_make_transport_stream_op(
 
 namespace grpc_core {
 
-ServerMetadataHandle ServerMetadataFromStatus(const absl::Status& status) {
-  auto hdl =
-      GetContext<Arena>()->MakePooled<ServerMetadata>(GetContext<Arena>());
+ServerMetadataHandle ServerMetadataFromStatus(const absl::Status& status,
+                                              Arena* arena) {
+  auto hdl = arena->MakePooled<ServerMetadata>(arena);
   hdl->Set(GrpcStatusMetadata(), static_cast<grpc_status_code>(status.code()));
   if (!status.ok()) {
     hdl->Set(GrpcMessageMetadata(), Slice::FromCopiedString(status.message()));
